/*
 *  <Short Description>
 *  Copyright (C) 2023  Brett Terpstra
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <blt/gfx/shader.h>
#include <blt/std/loader.h>
#include <blt/std/string.h>
#include <blt/std/memory.h>

namespace blt::gfx
{
    static std::string removeEmptyFirstLines(const std::string& string){
        auto lines = blt::string::split(string, "\n");
        std::string new_source_string;
        for (const auto& line : lines) {
            if (!line.empty() && !blt::string::contains(line, "\"")) {
                new_source_string += line;
                new_source_string += "\n";
            }
        }
        return new_source_string;
    }
    
    unsigned int shader_t::createShader(const std::string& source, int type) {
        const char* shader_code = source.c_str();
        // creates a Shader
        unsigned int shaderID = glCreateShader(type);
        // loads the shader code for later complication and uploading into the graphics card
        // TODO: defines can be added here by sending them as additional strings. No need to edit the source string
        glShaderSource(shaderID, 1, &shader_code, nullptr);
        // Compile it
        glCompileShader(shaderID);
        
        // make sure there are no errors in the compilation. If there is then print out information pertaining to the error.
        // the actual log is highly dependent on the platform this is being run from, so we cannot make any assumptions about the issue.
        // the TODO: maybe find a way of lexing the output to give suggestions about fixing the error? default error messages can be unhelpful at times.
        GLint success;
        glGetShaderiv(shaderID, GL_COMPILE_STATUS, &success);
        if (!success) {
            int log_length = 0;
            glGetShaderiv(shaderID, GL_INFO_LOG_LENGTH, &log_length);
            
            // scoped buffers will delete their memory when they go out of scope. A benefit of using BLT
            blt::scoped_buffer<GLchar> infoLog{static_cast<unsigned long>(log_length + 1)};
            
            glGetShaderInfoLog(shaderID, log_length + 1, nullptr, infoLog.data());
            auto shader_type_str = (type == GL_VERTEX_SHADER ? "Vertex Shader" : type == GL_FRAGMENT_SHADER ? "Fragment Shader" : "Other Shader");
            BLT_ERROR("--- --- --- --- --- --- --- --- ---");
            BLT_ERROR("Unable to compile shader of type %s\nShader source:", shader_type_str);
            BLT_ERROR(source);
            BLT_ERROR("I have an log of %d length", log_length);
            BLT_ERROR(infoLog.data());
            BLT_ERROR("--- --- --- --- --- --- --- --- ---");
        }
        return shaderID;
    }
    
    shader_t::shader_t(const std::string& vertex, const std::string& fragment, bool load_as_string) {
// load shader sources
        std::string vertex_source = vertex;
        std::string fragment_source = fragment;
        if (!load_as_string){
            // BLT provides a recursive file loader for glsl shaders. It's pretty much just a recursive function looking for include statements.
            vertex_source = blt::fs::loadShaderFile(vertex);
            fragment_source = blt::fs::loadShaderFile(fragment);
        } else {
            vertex_source = removeEmptyFirstLines(vertex_source);
            fragment_source = removeEmptyFirstLines(fragment_source);
        }
        
        // create the shaders
        vertexShaderID = createShader(vertex_source, GL_VERTEX_SHADER);
        fragmentShaderID = createShader(fragment_source, GL_FRAGMENT_SHADER);
        
        // bind them to a program
        programID = glCreateProgram();
        // attach the loaded shaders to the Shader program
        glAttachShader(programID, vertexShaderID);
        glAttachShader(programID, fragmentShaderID);
        // link and make sure that our program is valid.
        glLinkProgram(programID);
        
        GLint success;
        glGetProgramiv(programID, GL_LINK_STATUS, &success);
        if (!success) {
            int log_length = 0;
            glGetProgramiv(programID, GL_INFO_LOG_LENGTH, &log_length);
            
            // scoped buffers will delete their memory when they go out of scope.
            blt::scoped_buffer<GLchar> infoLog{static_cast<unsigned long>(log_length + 1)};
            
            glGetProgramInfoLog(programID, log_length + 1, nullptr, infoLog.data());
            BLT_ERROR("--- --- --- --- --- --- --- --- ---");
            BLT_ERROR("Unable to link program of ID: %d", programID);
            BLT_ERROR(vertex_source);
            BLT_ERROR(fragment_source);
            BLT_ERROR("I have an log of %d length", log_length);
            BLT_ERROR(infoLog.data());
            BLT_ERROR("--- --- --- --- --- --- --- --- ---");
        }
        
        glValidateProgram(programID);
        bind();
        setUniformBlockLocation("GlobalMatrices", 0);
        glUseProgram(0);
    }
    
    void shader_t::bindAttribute(int attribute, const std::string &name) const {
        bind();
        glBindAttribLocation(programID, attribute, name.c_str());
    }
    
    void shader_t::setUniformBlockLocation(const std::string &name, int location) const {
        bind();
        auto blockID = glGetUniformBlockIndex(programID, name.c_str());
        if (blockID != GL_INVALID_INDEX)
            glUniformBlockBinding(programID, blockID, location);
        else
            BLT_WARN("Unable to find uniform buffer '%s'. Did you forget to declare it?", name.c_str());
    }
    
    shader_t::~shader_t() {
        glUseProgram(0);
        // shader was moved
        if (programID <= 0)
            return;
        // remove all the shaders from the program
        glDetachShader(programID, vertexShaderID);
        glDetachShader(programID, fragmentShaderID);
        
        // delete the shaders
        glDeleteShader(vertexShaderID);
        glDeleteShader(fragmentShaderID);
        
        // delete the Shader program
        glDeleteProgram(programID);
    }
    
    shader_t::shader_t(shader_t&& move) noexcept {
        // the move constructor doesn't need to construct a new shader but it does need to ensure all old variables are moved over
        programID = move.programID;
        vertexShaderID = move.vertexShaderID;
        fragmentShaderID = move.fragmentShaderID;
        for (const auto& pair : move.uniformVars)
            uniformVars.insert(pair);
        // by setting the program ID to -1 we tell the shader it has been moved.
        move.programID = -1;
    }
    
    uniform_buffer::uniform_buffer(size_t size, GLuint location): size_(size), location_(location)
    {
        glGenBuffers(1, &uboID);
        bind();
        glBufferData(GL_UNIFORM_BUFFER, static_cast<GLsizeiptr>(size), nullptr, GL_DYNAMIC_READ);
        unbind();
        
        glBindBufferBase(GL_UNIFORM_BUFFER, location, uboID);
    }
    
    uniform_buffer::uniform_buffer(void* data, size_t size, GLuint location): size_(size), location_(location)
    {
        glGenBuffers(1, &uboID);
        bind();
        glBufferData(GL_UNIFORM_BUFFER, static_cast<GLsizeiptr>(size), data, GL_DYNAMIC_READ);
        unbind();
        
        glBindBufferBase(GL_UNIFORM_BUFFER, location, uboID);
    }
    
    uniform_buffer& uniform_buffer::resize(size_t newSize)
    {
        bind();
        size_ = newSize;
        glBufferData(GL_UNIFORM_BUFFER, static_cast<GLsizeiptr>(size_), nullptr, GL_DYNAMIC_READ);
        unbind();
        return *this;
    }
    
    uniform_buffer& uniform_buffer::upload(void* data, size_t size, size_t offset)
    {
        bind();
        glBufferSubData(GL_UNIFORM_BUFFER, static_cast<GLsizeiptr>(offset), static_cast<GLsizeiptr>(size), data);
        unbind();
        return *this;
    }
    
    uniform_buffer& uniform_buffer::bind()
    {
        glBindBuffer(GL_UNIFORM_BUFFER, uboID);
        return *this;
    }
    
    uniform_buffer::~uniform_buffer()
    {
        glDeleteBuffers(1, &uboID);
    }
}
